mod bytes;
mod crypto;
mod io;

use crate::bytes::Bytes;
use openssl::symm::Mode;
use rand::Rng;

fn q01() {
    println!("Running q01");

    let input = "49276d206b696c6c696e6720796f757220627261696e206c696b65206120706f69736f6e6f7573206d757368726f6f6d";
    let output = Bytes::from_hex(input).to_base64();
    let expected = String::from("SSdtIGtpbGxpbmcgeW91ciBicmFpbiBsaWtlIGEgcG9pc29ub3VzIG11c2hyb29t");
    assert!(expected == output);
}

fn q02() {
    println!("Running q02");
    let input1 = Bytes::from_hex("1c0111001f010100061a024b53535009181c");
    let input2 = Bytes::from_hex("686974207468652062756c6c277320657965");

    let expected = Bytes::from_hex("746865206b696420646f6e277420706c6179");
    let xor = input1.xor_fixed(&input2).unwrap();

    assert!(expected == xor);
}

fn q03() {
    println!("Running q03");
    let input =
        Bytes::from_hex("1b37373331363f78151b7f2b783431333d78397828372d363c78373e783a393b3736");
    let (best_cipher, _) = input.find_best_single_xor_cipher();
    let output = input.xor_byte(best_cipher);
    println!(
        "  0x{:X}: {}",
        best_cipher,
        String::from_utf8(output.data).unwrap()
    );
}

fn q04() {
    println!("Running q04");
    let mut best_line = Bytes::new();
    let mut best_score = 0.0f32;
    let mut best_cipher = 0u8;

    for line in io::read_lines("data/4.txt").unwrap() {
        if let Ok(line) = line {
            let line = Bytes::from_hex(&line);
            let (cipher, score) = line.find_best_single_xor_cipher();
            if score > best_score {
                best_score = score;
                best_cipher = cipher;
                best_line = line;
            }
        }
    }

    println!(
        "  0x{:2X} {}",
        best_cipher,
        String::from_utf8(best_line.xor_byte(best_cipher).data).unwrap()
    );
}

fn q05() {
    println!("Running q05");
    let input = Bytes::from_string(
        "Burning 'em, if you ain't quick and nimble\nI go crazy when I hear a cymbal",
    );
    let cipher = Bytes::from_string("ICE");
    let expected = Bytes::from_hex("0b3637272a2b2e63622c2e69692a23693a2a3c6324202d623d63343c2a26226324272765272a282b2f20430a652e2c652a3124333a653e2b2027630c692b20283165286326302e27282f");

    let output = input.xor_repeating(&cipher);

    assert!(expected == output);
}

fn q06() {
    println!("Running q06");

    assert!(
        37u32
            == Bytes::from_string("this is a test")
                .hamming_distance(&Bytes::from_string("wokka wokka!!!"))
    );

    let input = io::read_without_lines("data/6.txt").unwrap();
    let input = Bytes::from_base64(&input);

    let key = input.find_best_xor_cipher();
    let _output = input.xor_repeating(&key).to_string().unwrap();
    // println!("  {}", _output.replace("\n", "\n  "));
}

fn q07() {
    println!("Running q07");

    let input = io::read_without_lines("data/7.txt").unwrap();
    let input = Bytes::from_base64(&input);
    let key = Bytes::from_string("YELLOW SUBMARINE");
    let _output = Bytes {
        data: crypto::Aes::new(&key, false)
            .unwrap()
            .ecb(input.into(), Mode::Decrypt)
            .unwrap(),
    };
    // println!("  {}", _output.to_string().unwrap().replace("\n", "\n  "));
}

fn q08() {
    println!("Running q08");
    let input = io::read_lines("data/8.txt");

    let line_chunks: Vec<Vec<Box<[u8]>>> = input
        .unwrap()
        .map(|s| {
            hex::decode(s.unwrap())
                .unwrap()
                .chunks_exact(16)
                .map(|c| Box::from(c))
                .collect()
        })
        .collect();
    let line_repetitions = line_chunks.into_iter().map(|mut ch| {
        ch.sort_unstable();
        ch.iter()
            .fold((0u32, None), |(n, last), curr| {
                (n + (if Some(curr) == last { 1 } else { 0 }), Some(curr))
            })
            .0
    });
    let best_line = line_repetitions
        .enumerate()
        .max_by_key(|(_, num)| *num)
        .unwrap()
        .0;
    println!("  Line {} is AES-ECB", best_line);
}

fn q09() {
    println!("Running q09");
    let input = Bytes::from_string("YELLOW SUBMARINE");
    let len = 20 as usize;
    let output = input.pad_to_multiple(len);
    assert!(Bytes::from_string("YELLOW SUBMARINE\x04\x04\x04\x04") == output);
}

fn q10() {
    println!("Running q10");
    let input = io::read_without_lines("data/10.txt").unwrap();
    let input = Bytes::from_base64(&input);
    let key = b"YELLOW SUBMARINE";
    let iv = [0u8; 16];

    let _output = crypto::Aes::new(key, false)
        .unwrap()
        .cbc(input.into(), &iv, Mode::Decrypt);
    // println!("{}", Bytes::from(_output.unwrap()).to_string().unwrap().replace("\n", "\n  "));
}

fn q11() {
    println!("Running q11");
    let black_box = |mut data: Vec<u8>| -> (bool, Vec<u8>) {
        let mut rng = rand::thread_rng();
        let dist = rand::distributions::Standard;
        let key: Vec<u8> = rng.sample_iter(&dist).take(16).collect();
        let iv: Vec<u8> = rng.sample_iter(&dist).take(16).collect();
        let aes = crypto::Aes::new(&key, true).unwrap();
        let mut start: Vec<u8> = rng.sample_iter(&dist).take(rng.gen_range(5,11)).collect();
        let mut end: Vec<u8> = rng.sample_iter(&dist).take(rng.gen_range(5,11)).collect();

        start.append(&mut data);
        start.append(&mut end);

        match rand::random() {
            false => (false, aes.cbc(start, &iv, Mode::Encrypt).unwrap()),
            true => (true, aes.ecb(start, Mode::Encrypt).unwrap()),
        }
    };

    let oracle = |data: &Vec<u8>| -> bool {
        data[16..32] == data[32..48]
    };

    for i in 0..128 {
        let data = [0u8;64];
        let (ecb, data) = black_box(data.to_vec());
        assert!(ecb == oracle(&data));
    }
}

fn q12()
{
    println!("Running q12");
    let key: Vec<u8> = rand::thread_rng().sample_iter(rand::distributions::Standard).take(16).collect();
    let plaintext: Vec<u8> = Bytes::from_base64("Um9sbGluJyBpbiBteSA1LjAKV2l0aCBteSByYWctdG9wIGRvd24gc28gbXkgaGFpciBjYW4gYmxvdwpUaGUgZ2lybGllcyBvbiBzdGFuZGJ5IHdhdmluZyBqdXN0IHRvIHNheSBoaQpEaWQgeW91IHN0b3A/IE5vLCBJIGp1c3QgZHJvdmUgYnkK").into();
    let aes = crate::crypto::Aes::new(&key, true).unwrap();

    let blackbox = |mut data: Vec<u8>| -> Vec<u8> {
        data.append(&mut plaintext.clone());
        aes.ecb(data, Mode::Encrypt).unwrap()
    };

    let starting_size = blackbox(Vec::new()).len();
    let (first_increment, next_size) = (0..).map(|x:usize| (x, blackbox(std::iter::repeat(0u8).take(x).collect()).len())).find(|(_,x)| *x != starting_size).unwrap();
    let block_size = next_size - starting_size;
    println!("  Block size is {}", block_size);
    assert!(block_size == 16);

    let long_block = blackbox(std::iter::repeat(0u8).take(128).collect());
    let using_ecb = long_block[48..64] == long_block[64..80];
    println!("  Using ECB: {}", using_ecb);
    assert!(using_ecb);

    println!("  First size increment at {}", first_increment);

    let data_length = starting_size - first_increment;
    let mut known: Vec<u8> = Vec::with_capacity(data_length);
    println!("  Data length: {}", data_length);

    for _ in 0..data_length
    {
        let nulls: Vec<u8> = std::iter::repeat(0u8).take(starting_size - known.len() - 1).collect();
        let target_block = &blackbox(nulls.clone())[0..starting_size];
        let mut found = false;
        'byteiter: for b in 0u8..=255u8 {
            let this_block = &blackbox(nulls.iter().chain(known.iter()).chain(std::iter::once(&b)).cloned().take(starting_size).collect())[0..starting_size];
            if this_block == target_block
            {
                known.push(b);
                found = true;
                break 'byteiter;
            }
        }
        assert!(found);
    }
    // println!("  {}", String::from_utf8(known).unwrap().replace("\n", "\n  "));
}

fn q13()
{
    println!("Running q13");
    let key: Vec<u8> = rand::thread_rng().sample_iter(rand::distributions::Standard).take(16).collect();

    let parse = |x: &str| -> std::collections::HashMap<String, String> {
        let mut out = std::collections::HashMap::new();
        for pair in x.split(|y| y == '&') {
            let elements: Vec<&str> = pair.split(|z| z == '=').collect();
            if elements.len() == 2 {
                out.insert(elements[0].to_owned(), elements[1].to_owned());
            }
        };
        out
    };

    let profile_for = |mail: &str| -> String {
        format!("email={}&uid=10&role=user", mail.replace("&","").replace("=",""))
    };

    let encrypt = |profile: &str| -> Vec<u8> {
        let profile: Vec<u8> = profile.bytes().collect();
        let aes = crypto::Aes::new(&key, true).unwrap();
        aes.ecb(profile, Mode::Encrypt).unwrap()
    };

    let is_admin = |profile: Vec<u8>| -> bool {
        let aes = crypto::Aes::new(&key, true).unwrap();
        let profile = aes.ecb(profile, Mode::Decrypt).unwrap();
        let profile = String::from_utf8(profile).unwrap();
        parse(&profile)["role"] == "admin"
    };

    let mail1 = "a".repeat(32 - "email=".len() - "@example.com&uid=10&role=".len());
    let mail1 = mail1 + "@example.com";
    let part1: &[u8] = &encrypt(&profile_for(&mail1))[0..32];
    let mail2 = " ".repeat(16 - "email=".len());
    let mail2 = mail2 + "admin";
    let padding_char = (16 - "admin".len()) as u8;
    let mail2 = mail2 + &String::from_utf8(vec![padding_char]).unwrap().repeat(padding_char as usize);
    let part2: &[u8] = &encrypt(&profile_for(&mail2))[16..32];
    let whole: Vec<u8> = part1.iter().chain(part2.iter()).cloned().collect();

    assert!(is_admin(whole));
}

fn main() {
    q01();
    q02();
    q03();
    q04();
    q05();
    q06();
    q07();
    q08();
    q09();
    q10();
    q11();
    q12();
    q13();
}
